/* SIMPLE - Simple Is a Macro Processing Language Element */
/* Copyright (c) 1998 David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This file is the global header file, containing definitions that
 * are needed about everywhere and prototypes for public functions.
 */

/* Version identifiers */

#define PRODUCTSTRING "SIMPLE"
#define VERSIONSTRING "2.0.0"
#define DATESTRING "18JAN98"

/* Essential typedefs */

typedef int token;
/* We define the token to be an ordinal type because they are _much_
 * easier to handle (than structures, for example). */

typedef struct token_list {
  token *l; /* Pointer to the actual list of tokens. */
  /* This pointer is always malloc()ated, and realloc()ated when the
   * need arises. */
  int size; /* Size of the list (number of tokens in it). */
  int allsize; /* Allocated size (also in tokens, _not_ bytes). */
} token_list;

typedef int command;
/* Command numbers used to be tokens in version 1.x.x but the complete
 * change in syntax has made it otherwise. We use negative numbers for
 * builtin commands and positive numbers for user-defined macros. */

/* Token values */

/* The actual values are unimportant. However, the special tokens
 * should come after characters (because to test whether a token is
 * special we use tests such as `if (t>=FIRST_SPECIAL)' */

/* Note that token value -1 is reserved for EOF. No other negative
 * value is used. Also, value 256 is reserved (who knows, it may turn
 * out to be useful some day). */

#define FIRST_SPECIAL 256
/* If we had a version which works with unicode we would change that
 * to 65536, I guess. */

/* These are the special tokens. */
#define BEGIN_COMMAND (FIRST_SPECIAL+1) /* < */
#define END_COMMAND (FIRST_SPECIAL+2)   /* > */
#define NEXT_PARAM (FIRST_SPECIAL+3)    /* | */
#define OPEN_QUOTE (FIRST_SPECIAL+4)    /* [ */
#define CLOSE_QUOTE (FIRST_SPECIAL+5)   /* ] */
#define QUOTE_NEXT (FIRST_SPECIAL+6)    /* # */
#define AT_SIGN (FIRST_SPECIAL+7)       /* @ */

/* Command values */

/* Command value 0 is reserved (it is used to mark the end of a
 * dictionary). */

/* These are the command numbers of builtin functions */
#define BI_IDENTITY (-1)
#define BI_VOID (-2)
#define BI_OUTPUT (-3)
#define BI_DEFINE (-4)
#define BI_INCLUDE (-5)
#define BI_IT (-6)
#define BI_QUOTE (-7)
#define BI_IF (-8)
#define BI_HEAD (-9)
#define BI_TAIL (-10)
#define BI_AHEAD (-11)
#define BI_ATAIL (-12)
#define BI_PUSH (-13)
#define BI_LAST (-14)
#define BI_POP (-15)
#define BI_POPLAST (-16)
#define BI_DEPTH (-17)
#define BI_PLUS (-18)
#define BI_MINUS (-19)
#define BI_TIMES (-20)
#define BI_DIV (-21)
#define BI_MOD (-22)
#define BI_EQ (-23)
#define BI_NEQ (-24)
#define BI_GT (-25)
#define BI_GE (-26)
#define BI_LT (-27)
#define BI_LE (-28)
#define BI_AND (-29)
#define BI_OR (-30)
#define BI_NOT (-31)
#define BI_BAND (-32)
#define BI_BOR (-33)
#define BI_BNOT (-34)

void fatal(char internal,char *fmt, ...);

void pushinputfile(char *name);

token identifycommand(char *name);

token inputtoken(void);

void outputtoken(token tk);

void expandinput(void);

void callmacro(token_list *expansion,int argc,token_list *argv);

void callbuiltin(token_list *expansion,command cmd,int argc,token_list *argv);

void calluser(token_list *expansion,command cmd,int argc,token_list *argv);

void defineuser(token_list *definition,token tk);
