/* SIMPLE - Simple Is a Macro Processing Language Element */
/* Copyright (c) 1998 David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Handle user-defined macros (defining them, calling them, etc.).
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "simple.h"
#include "token.h"

typedef struct def_entry {
  command who; /* What command is being defined */
  token_list what; /* Definition of that token */
} def_entry;

static def_entry *user_table = NULL;
static int user_table_len = 0;

void
replaceuser(int entry,token_list *definition)
     /* Modify the definition of a user token */
{
  freelist(&user_table[entry].what);
  copylist(&user_table[entry].what,definition);
}

void
createuser(token_list *definition,command cmd)
     /* Create a new user token */
{
  if ((user_table=realloc(user_table,(++user_table_len)*sizeof(def_entry)))
      ==NULL)
    fatal(0,"Out of memory!");
  user_table[user_table_len-1].who=cmd;
  makeemptylist(&user_table[user_table_len-1].what);
  replaceuser(user_table_len-1,definition);
}

void
defineuser(token_list *definition,command cmd)
     /* Define a user token, checking to see whether it already exists */
{
  int i;
  for (i=0;i<user_table_len;i++) {
    if (cmd==user_table[i].who) {
      replaceuser(i,definition);
      return;
    }
  }
  createuser(definition,cmd);
}

void
expanduser(token_list *expansion,token_list *definition,
	   int argc,token_list *argv)
     /* Expand user command definition, putting the result in
      * expansion, the number of arguments being argc and their values
      * given by argv. */
{
  int i,j;
  token_list tmplist;
  char *tmpstring,*tmpstring2;
  int cnt,cnt2;
  for (i=0;i<definition->size;i++) {
    if (definition->l[i]!=AT_SIGN)
      appendtoken(expansion,definition->l[i]);
    else {
      makeemptylist(&tmplist);
      for (i++;(i<definition->size)&&(definition->l[i]!=AT_SIGN);i++)
	appendtoken(&tmplist,definition->l[i]);
      if (i==definition->size)
	fatal(0,"Bad parameter construct (missing @).");
      tmpstring=tokenlisttostring(&tmplist);
      freelist(&tmplist);
      if (strcmp(tmpstring,"")==0) {
	appendtoken(expansion,AT_SIGN); /* Whatever that's good for... */
      } else if (strcmp(tmpstring,"?")==0) {
	tmpstring2=malloc(30);
	if (snprintf(tmpstring2,30,"%d",argc-1)==30)
	  fatal(1,"Ooops! not enough characters...");
	for (j=0;tmpstring2[j];j++)
	  appendtoken(expansion,tmpstring2[j]);
	free(tmpstring2);
      } else if (tmpstring[0]=='.') {
	if (strcmp(tmpstring,".")==0)
	  cnt=1;
	else if (sscanf(tmpstring+1,"%d",&cnt)<1)
	  fatal(0,"Bad parameter construct (didn't understand %s).",tmpstring);
	if (cnt<0) cnt+=argc;
	for (cnt2=cnt;(cnt2>=0)&&(cnt2<argc);cnt2++) {
	  appendtoken(expansion,NEXT_PARAM);
	  appendlist(expansion,argv+cnt2);
	}
      } else if (tmpstring[0]==';') {
	if (strcmp(tmpstring,";")==0)
	  cnt=1;
	else if (sscanf(tmpstring+1,"%d",&cnt)<1)
	  fatal(0,"Bad parameter construct (didn't understand %s).",tmpstring);
	if (cnt<0) cnt+=argc;
	for (cnt2=cnt;(cnt2>=0)&&(cnt2<argc);cnt2++) {
	  appendtoken(expansion,NEXT_PARAM);
	  for (j=0;j<argv[cnt2].size;j++) {
	    appendtoken(expansion,QUOTE_NEXT);
	    appendtoken(expansion,argv[cnt2].l[j]);
	  }
	}
      } else if (tmpstring[0]==',') {
	if (sscanf(tmpstring+1,"%d",&cnt)<1)
	  fatal(0,"Bad parameter construct (didn't understand %s).",tmpstring);
	if (cnt<0) cnt+=argc;
	if ((cnt>=0)&&(cnt<argc)) {
	  for (j=0;j<argv[cnt].size;j++) {
	    appendtoken(expansion,QUOTE_NEXT);
	    appendtoken(expansion,argv[cnt].l[j]);
	  }
	}
      } else {
	if (sscanf(tmpstring,"%d",&cnt)<1)
	  fatal(0,"Bad parameter construct (didn't understand %s).",tmpstring);
	if (cnt<0) cnt+=argc;
	if ((cnt>=0)&&(cnt<argc))
	  appendlist(expansion,argv+cnt);
	else /* Append the empty list: no-op. */;
      }
      free(tmpstring);
    }
  }
}

void
calluser(token_list *expansion,command cmd,int argc,token_list *argv)
     /* Expand user command cmd looking up its definition, and put the
      * result in expansion, the number of arguments being argc and
      * their values given by argv. */
{
  int i;
  for (i=0;i<user_table_len;i++) {
    if (cmd==user_table[i].who) {
      expanduser(expansion,&user_table[i].what,argc,argv);
      return;
    }
  }
  fatal(0,"Undefined command.");
}
